/***************************************************************************
 * 
 * Copyright 1995  by  the  Massachusetts Institute  of Technology.   All  
 * rights reserved.
 * 
 * Developed  by the Perceptual Computing Section 
 * at the Media Laboratory, MIT, Cambridge, Massachusetts.
 * 
 * Permission to use, copy,  or modify this software and  its documentation
 * for  educational  and  research purposes only and without fee  is hereby
 * granted, provided  that this copyright notice and the original authors's
 * names appear  on all copies and supporting documentation.  If individual
 * files are  separated from  this  distribution directory  structure, this
 * copyright notice must be included.  For any other uses of this software,
 * in original or  modified form, including but not limited to distribution
 * in whole or in  part, specific  prior permission  must be  obtained from
 * MIT.  These programs shall not  be  used, rewritten, or  adapted as  the
 *   basis  of  a commercial  software  or  hardware product  without first
 * obtaining appropriate licenses  from MIT.  MIT. makes no representations
 * about the suitability of this  software for any purpose.  It is provided
 * "as is" without express or implied warranty.
 * 
 **************************************************************************/

// pfCamera.c++

#include "pfCamera.h++"

///////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////

BackProject::BackProject(SoPerspectiveCamera *cam)
{
    xf = new SoTransform();
    xf->rotation.setValue(cam->orientation.getValue());
    cp = cam->position.getValue();
    xf->translation.setValue(cp[0],cp[1],0);

    SoGetMatrixAction *gma = new SoGetMatrixAction(SbViewportRegion());
    gma->apply(xf);
    matrix = gma->getMatrix();
    m2 = gma->getInverse();
    delete gma;
    
    // sm is a matrix which shifts the screen point (accounting for
    // 0 to 1 viewport mapping, and positions it in z. Currently,
    // the z is a fudge factor that depends on the camera parameters
    
    spm_static.makeIdentity();
    spm_static.setTranslate(SbVec3f(-.5, -.5, -1.2*.78/(cam->heightAngle.getValue())));
    spm1.makeIdentity();
    spm1.setScale(SbVec3f(4.0/3.0, 1, 1));
    spm_static.multRight(spm1);
    planeZ = 0;
}

void BackProject::setProjectionPlane(float z)
{
    plane = SbPlane(SbVec3f(0, 0, 1), -z+cp[2]);
    plane.transform(m2);
    SbVec3f norm = plane.getNormal();
    float d = plane.getDistanceFromOrigin();
    planeZ = z;
    //Now build the projection matrix
    proj = SbMatrix(
	-d, 0,  0,  norm[0],
	0, -d,  0,  norm[1],
	0, 0,  -d,  norm[2],
	0, 0,  0, 0); 
    
    // Now concatenate up the matrix
    // sm * proj * matrix
    // The first guy converts the point to the camera coordinates from
    // the screen coordinates. The second guy projects it onto the projection
    // plane which has been transformed into the camera's frame. The third
    // guy takes the result and converts it back into the world coordinates
    // spm can be calculated once per projection plane
    spm = spm_static;
    spm.multRight(proj);
    spm.multRight(matrix);
    
    SbMatrix spfinal;
    spfinal.identity();
    spfinal.setTranslate(SbVec3f(0, 0, cp[2]));
    spm.multRight(spfinal);
}

void BackProject::backProjectPt(const SbVec3f &sp1, SbVec3f &projPt, float z)
{
    if(fabs(z-planeZ)> .1)
	setProjectionPlane(z);
    SbVec3f sp = sp1;
    sp[2] = 0;
    // This does the actual conversion
    spm.multVecMatrix(sp, projPt);
    //printf("%f %f %f \n", projPt[0], projPt[1], projPt[2]);
    return;
}

///////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////

static
inline
void
error( char *fn, char *msg, int doexit )
{
    cerr << "pfCamera::" << fn << ": " << msg << endl;
    if( doexit )
       exit( 1 );
}

pfCamera::pfCamera( char *filename )
  : camera( new SoPerspectiveCamera ),
    projector( 0 ),
    heightToWidth( 0. )
{
    float sx;
    float sy;
    float dp[3];
    float mat[3][3];
    float ha;
    float sc;

    // (1) READ FILE
    ifstream file( filename );
    if( !file )
       error( "pfCamera", "camera input file not found", 1 );
    file >> sx;				// screen dimensions
    file >> sy;
    for( int j = 0; j < 3; j++ )	// position of camera
       file >> dp[j];
    for( int i = 0; i < 3; i++ )	// rotation matrix
       for( j = 0; j < 3; j++ )
	  file >> mat[i][j];
    file >> ha;				// height angle
    file >> sc;				// scale
    file.close();

    // (2) ADJUST CAMERA
    SbMatrix rm( mat[0][0], mat[0][1], mat[0][2], 0, 
		 mat[1][0], mat[1][1], mat[1][2], 0,
		 mat[2][0], mat[2][1], mat[2][2], 0,
		 0, 0, 0, 1);
    SbVec3f disp( dp[0], dp[1], dp[2] ); 
    disp /= sc;
    camera -> aspectRatio.setValue( SO_ASPECT_VIDEO );
    heightToWidth = 1. / SO_ASPECT_VIDEO;
    camera -> viewportMapping.setValue( SoCamera::CROP_VIEWPORT_FILL_FRAME );
    camera -> position.setValue( disp );
    camera -> orientation.setValue( SbRotation( rm ) );
    camera -> nearDistance = 1.;
    camera -> farDistance  = 1000.;
    camera -> heightAngle = ha;
    camera -> ref();
    projector = new BackProject( camera );
    if( !projector )
       error( "pfCamera", "creation of BackProject failed", 1 );
}

void
pfCamera::projectToScreen( const float in[3], float out[3] )
{
    SbVec3f outvec;
    camera -> getViewVolume().projectToScreen( SbVec3f( in ), outvec );
    outvec.getValue( out[0], out[1], out[2] );
    out[1] = 1. - out[1];
}

void
pfCamera::projectToScreen2( const float in[3], float out[2] )
{
    SbVec3f outvec;
    float dummy;
    camera -> getViewVolume().projectToScreen( SbVec3f( in ), outvec );
    outvec.getValue( out[0], out[1], dummy );
    out[1] = 1. - out[1];
}

void
pfCamera::projectYtoZFloor( const float in[2], float out[3] )
{
    SbVec2f screenPt( in[0], 1. - in[1] );
    // SbVec2f screenPt( in[0], in[1] );
    SbVec3f worldPt;
    SbPlaneProjector pp( SbPlane( SbVec3f(0, 1, 0), 0), 0);
    pp.setViewVolume( camera -> getViewVolume() );
    worldPt = pp.project( screenPt );
    worldPt.getValue( out[0], out[1], out[2] );
}
    
void
pfCamera::projectScreenYtoWorldY( const float in[2], float out[3], float z )
{
    SbVec3f screenPt( in[0], 1.-in[1], 0 );
    SbVec3f worldPt;
    projector -> backProjectPt( screenPt, worldPt, z );
    worldPt.getValue( out[0], out[1], out[2] );
}

int
pfCamera::projectScreenYtoWorldY( int n, float* in, float* out, float z )
{
    // "in" is an array of 2-D points
    // "out" is an array of 3-D points
    for( int i = 0; i < n; i++ )
       projectScreenYtoWorldY( in+2*i, out+3*i, z );
    return i;
}
