/***************************************************************************
 * 
 * Copyright 1995  by  the  Massachusetts Institute  of Technology.   All  
 * rights reserved.
 * 
 * Developed  by the Perceptual Computing Section 
 * at the Media Laboratory, MIT, Cambridge, Massachusetts.
 * 
 * Permission to use, copy,  or modify this software and  its documentation
 * for  educational  and  research purposes only and without fee  is hereby
 * granted, provided  that this copyright notice and the original authors's
 * names appear  on all copies and supporting documentation.  If individual
 * files are  separated from  this  distribution directory  structure, this
 * copyright notice must be included.  For any other uses of this software,
 * in original or  modified form, including but not limited to distribution
 * in whole or in  part, specific  prior permission  must be  obtained from
 * MIT.  These programs shall not  be  used, rewritten, or  adapted as  the
 *   basis  of  a commercial  software  or  hardware product  without first
 * obtaining appropriate licenses  from MIT.  MIT. makes no representations
 * about the suitability of this  software for any purpose.  It is provided
 * "as is" without express or implied warranty.
 * 
 **************************************************************************/

// pfFeatures.h++

#ifndef _pfFeatures_H_
#define _pfFeatures_H_

#include <stream.h>
#include <Inventor/SbLinear.h>
#include <Inventor/nodes/SoCoordinate3.h>
#include "pfCamera.h++"
#include "pfSemaphore.h++"
#include "pfContour.h++"

extern "C"
{
#include "bsColor.h"
#include "bsClassify.h"
}

#ifndef MAX
#define MAX(a,b) (((a) > (b)) ? (a) : (b))
#define MIN(a,b) (((a) < (b)) ? (a) : (b))
#endif

enum boolean { pfFALSE = 0, pfTRUE = 1 };

class ContourPointSet
{
  const int maxList = 2000;
public:
  int n;
  float* points[ maxList ];
  ContourPointSet() : n(0) {}
  int add( float* fp ) { points[n++] = fp; return n; }
};

const int MAX_COLORCLASSES = 50;
class ColorClass2DInfo
{
public:
  float cmean[3];	        /* color rgb */
  float gmean[2];	        /* screen xy */
  float gL[2];			/* eigen values */
  float gS[2][2];		/* eigen vectors */
};

class ColorClass3DInfo
{
public:
  float cmean[3];	        /* color rgb */
  float gmean[3];	        /* screen xy */
  float gL[3];			/* eigen values */
  float gS[3][3];		/* eigen vectors */
};

typedef struct
{
  float imageCentroid[2];
  float imageBoxMin[2];
  float imageBoxMax[2];
  float imageHead[2];
  float imageLeftHand[2];
  float imageRightHand[2];
  float imageLeftFoot[2];
  float imageRightFoot[2];
  float userplaneCentroid[3];
  float userplaneBoxUL[3];
  float userplaneBoxBR[3];
  float userplaneHead[3];
  float userplaneLeftHand[3];
  float userplaneRightHand[3];
  float userplaneLeftFoot[3];
  float userplaneRightFoot[3];
  boolean headValid;
  boolean leftHandValid;
  boolean rightHandValid;
  boolean leftFootValid;
  boolean rightFootValid;
  int numContourPoints;
  float imageContour[ pfContour::maxContourLength * 2 ];
  float userplaneContour[ pfContour::maxContourLength * 3 ];
  int numColorClasses;
  ColorClass2DInfo colorclasses2d[ MAX_COLORCLASSES ];
  ColorClass3DInfo colorclasses3d[ MAX_COLORCLASSES ];
  float userheight;
} _feature_data;

class pfFeatures
{
    _feature_data  featureBuffers[2];
    _feature_data* features;
    _feature_data* featuresLatest;
    int numExtremePoints;
    float scale;// camera to world 
    const int minContourLength = 40;
public:
    pfSemaphore semaphore;
    pfFeatures();
    inline void setScale( float sc ) { scale = sc; }
    inline _feature_data& getLatest() { return *featuresLatest; }
    inline _feature_data& getCurrent() { return *features; }
    inline void swapBuffers();
    int ProjectPfContourToUserPlane( pfContour&, pfCamera* );
    void PFind( _feature_data*, pfCamera* );
    void Find( pfCamera* ); // camera used for getting the 2D features back
    inline void SetNotValid();
    ostream& Print( ostream& );
    ostream& Print2d( ostream& );
    static float hmin_pu;
    static float hmax_pu;
    static void setHandMin( float x ) { hmin_pu = x; }
    static void setHandMax( float x ) { hmax_pu = x; }
    static float getHandMin() { return hmin_pu; }
    static float getHandMax() { return hmax_pu; }
    static float feetTol;
    static float headTol;
    static float extendedTol;
    static float xLeft;
};

inline
void
pfFeatures::swapBuffers()
{
    semaphore.localLOCK();

    _feature_data *tmp = features;
    features = featuresLatest;
    featuresLatest = tmp;

    semaphore.localUNLOCK();
}

inline
void
pfFeatures::SetNotValid()
{
    features -> headValid = pfFALSE;
    features -> leftHandValid = pfFALSE;
    features -> rightHandValid = pfFALSE;
    features -> leftFootValid = pfFALSE;
    features -> rightFootValid = pfFALSE;
    features -> numContourPoints = 0;
}

inline
void
vectorSetValue( float* v, float x, float y, float z )
{
    v[0] = x;
    v[1] = y;
    v[2] = z;
}

inline
void
vectorSetValue( float* v, float x, float y )
{
    v[0] = x;
    v[1] = y;
}

inline
void
vector3Copy( float* in, float* out )
{
    bcopy( in, out, 3*sizeof(float) );
}

inline
void
vector3SubtractEq( float* out, float* v )
{
    out[0] -= v[0];
    out[1] -= v[1];
    out[2] -= v[2];
}

inline
ostream&
_print_vector( ostream& os, float* v )
{
    return( os << v[0] << ", " << v[1] << ", " << v[2] );
}

inline
ostream&
_print_vector2( ostream& os, float* v )
{
    return( os << v[0] << ", " << v[1] );
}

extern int handMin;
extern int handMax;

#endif /* _pfFeatures_H_ */

